#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include "tetX.h"
#include "keyjoy.h"

extern Display	*display;
extern Visual   *visual;
extern Window   window;
extern Pixmap   screenContext;
extern GC	graphContext;

extern int	ShowGrid;
extern int	ScreenDepth;
extern int	screenNumber;
extern int	TimerSuspended;

extern int	JoyEX, JoyEY, JoyW, JoyH;
extern Bool	DoInvalidate;

extern int	Demo;
int ShadowY;

#define JOYARR_HEIGHT  14
#define JOYARR_WIDTH   16
#define JOYARR_WIDTHB	8
#define JOYARR_COUNT	5

static XImage *PatIm[NPATS]; 
static XImage *JoyArrActiveIm[JOYARR_COUNT], *JoyArrPassiveIm[JOYARR_COUNT]; 

extern unsigned long PixValues[16];
extern int   BoardX, BoardY, BoardW, BoardH;
extern char  Patterns[NPATS][FIG_HEIGHT][FGB_WIDTH];
extern char  ArrowsActive[5][14][8], ArrowsPassive[5][14][8];
extern int   ArrOffs[5][2];
//-------------------------------------------------------------
int InitPixmapPats(void)
{ 
    XImage	*xi, *xi1;
    unsigned   	patno;
    unsigned 	x, y, x1;
    unsigned	c, c1;

    unsigned	char	*data;
    
/*   unsigned  char    *pat, *data;    
    unsigned  int     linelen, increm;
    unsigned  char    *target;
*/

    bzero(PatIm, NPATS * sizeof(XImage *));
    
                    

    for (patno=0; patno < NPATS; patno++)
    {
         xi = XCreateImage(display, visual, ScreenDepth, ZPixmap, 0,
	      NULL, FIG_WIDTH, FIG_HEIGHT,  8, 0);
        if (!xi)
	{ fprintf (stderr, "Failed to allocate pattern %d image header\n", patno);
	  return 1;
	}
	   

	data = malloc(FIG_HEIGHT * xi->bytes_per_line);
	if (!data)
	{ fprintf (stderr, "Failed to allocate pattern %d image\n", patno);
	  return 1;
	}
        xi->data = data;


/* This works faster with Linux gcc on Intel, but is not portable 
        linelen = xi->bytes_per_line;
	increm = ScreenDepth / 8;
    	pat = (char *) Patterns[patno];

        for (y = 0; y < FIG_HEIGHT; y++)
        { 
	    x1 = 0;
	    target = (data + y*linelen);
	    for (x = 0; x < FGB_WIDTH; x++)                  
            {             
	       if (ScreenDepth == 4) *target++ = *pat++;
	       else
	       {   c = *pat++;
	       
                   *(unsigned long *)target = PixValues[(c>>4) & 0xf];
                   target += increm;
		   
                   *(unsigned long *)target = PixValues[c & 0xf];
                   target += increm;
	       }   
	    }
	    
	}
*/


        for (y = 0; y < FIG_HEIGHT; y++)
        { 
	    x1 = 0;
	    for (x = 0; x < FGB_WIDTH; x++)                  
            {  
	    
	       c = Patterns[patno][y][x] & 0xff;
	       XPutPixel(xi, x1++, y, PixValues[c>>4]);
	       XPutPixel(xi, x1++, y, PixValues[c & 0xf]);
	    }
	    
	}


	
	PatIm[patno] = xi;
	
    }	
    

    /* Now do the same for joystick arrow images  */
    bzero(JoyArrActiveIm, JOYARR_COUNT * sizeof(XImage *));
    bzero(JoyArrPassiveIm, JOYARR_COUNT * sizeof(XImage *));
    
                    

    for (patno=0; patno < JOYARR_COUNT; patno++)
    {   /* For active arrow */
        xi = XCreateImage(display, visual, ScreenDepth, ZPixmap, 0,
	      NULL, JOYARR_WIDTH, JOYARR_HEIGHT,  8, 0);
        if (!xi)
	{ fprintf (stderr, "Failed to allocate joystick arrow %d image header\n", patno);
	  return 1;
	}
	   
	/* Allocate size */
	x = JOYARR_HEIGHT * xi->bytes_per_line;
	data = malloc(x);
	if (!data)
	{ fprintf (stderr, "Failed to allocate joystick arrow %d image\n", patno);
	  return 1;
	}
        xi->data = data;


	/* For passive arrow */
	data = malloc(x);
        xi1 = XCreateImage( display, visual, ScreenDepth, ZPixmap, 0,
	       data, JOYARR_WIDTH, JOYARR_HEIGHT,  8, 0);

        if (!xi1)
	{ fprintf (stderr, "Failed to allocate joystick arrow %d image header\n", patno);
	  return 1;
	}
	   

        for (y = 0; y < JOYARR_HEIGHT; y++)
        { 
	    x1 = 0;
	    for (x = 0; x < JOYARR_WIDTHB; x++)                  
            {  
	    
	       c = ArrowsActive[patno][y][x] & 0xff;
	       c1 = ArrowsPassive[patno][y][x] & 0xff;

	       XPutPixel(xi, x1, y, PixValues[c>>4]);
	       XPutPixel(xi1, x1++, y, PixValues[c1>>4]);

	       XPutPixel(xi, x1, y, PixValues[c & 0xf]);
	       XPutPixel(xi1, x1++, y, PixValues[c1 & 0xf]);

	    }
	    
	}

	
	JoyArrActiveIm[patno] = xi;
	JoyArrPassiveIm[patno] = xi1;
	
    }	
    return 0;	    
}

void FreePixmapPats(void)
{ 
    int       patno;
    XImage    *xi;

    for (patno=0; patno < NPATS; patno++)
    { xi = PatIm[patno];
      if (xi) XDestroyImage(xi);
    }  

    for (patno=0; patno < JOYARR_COUNT; patno++)
    { xi = JoyArrActiveIm[patno];
      if (xi) XDestroyImage(xi);

      xi = JoyArrPassiveIm[patno];
      if (xi) XDestroyImage(xi);
    }   
}


//-------------------------------------------------------------

void DrawBoardRow(char brdrow[BRD_WIDTH], int fx, int tx, int y, unsigned char mask)
{ int x, x2, y2;
  XGCValues  xgcv; 
  int patno;

  XImage *xim;
  long  size;
  unsigned  char  *data;
  

      y2 = BoardY + y  * FIG_HEIGHT;
      x2 = BoardX + fx * FIG_WIDTH;

      xgcv.plane_mask = -1;
//      if ( (unsigned char)mask == 0xff) xgcv.function = GXcopyInverted;
//      else
      xgcv.function = GXcopy;
      XChangeGC(display, graphContext, GCFunction | GCPlaneMask, &xgcv);
      
      if (ScreenDepth == 8) mask &= 7;

      for (x=fx; x < tx; x++)
      { patno = brdrow[x];
        if (ShowGrid && patno == 0) patno = 8;

        if (mask)
        {   xim = XSubImage(PatIm[patno], 0, 0, FIG_WIDTH, FIG_HEIGHT);
            size = (long)(xim->bytes_per_line) * FIG_HEIGHT;
	    data = (unsigned char *) (xim->data);
	    while (--size >= 0) *(data+size) ^= mask;
            XPutImage(display, screenContext, graphContext, 
	            xim, 0, 0, x2, y2, FIG_WIDTH, FIG_HEIGHT);
            XDestroyImage(xim);
        }
	else	    
        XPutImage(display, screenContext, graphContext, 
	          PatIm[patno], 0, 0, x2, y2, FIG_WIDTH, FIG_HEIGHT);


    
        x2 += FIG_WIDTH;		         
      }

      return;
}



//-------------------------------------------------------------

void DrawBoard(BOARD brd[][BRD_WIDTH],
          int fromx, int tox, int fromy, int toy)
{   int  y;
    RECT rc;

    if (fromy < 0) fromy = 0;
    if (toy > BRD_HEIGHT) toy = BRD_HEIGHT;

    if (fromx < 0) fromx = 0;
    if (tox > BRD_WIDTH) tox = BRD_WIDTH;

    for (y=fromy; y<toy; y++)
       DrawBoardRow(brd[y], fromx, tox,  y, 0);
       
       
    rc.left  = BoardX + fromx*FIG_WIDTH;
    rc.right = BoardX + tox * FIG_WIDTH;
     
    rc.top   = BoardY + fromy*FIG_HEIGHT;
    rc.bottom= BoardY + toy*FIG_HEIGHT;
    
    if (DoInvalidate) ContextRectToScreen(&rc);
        
     if (Demo) drawrobbrd(brd, fromx, tox, fromy, toy);

    

}


/*--------------------------------------------------------*/
void ShowFigure(FIGURE figure, int px, int py, int tox, int toy)
{
    int  x, y, x2, y2;
    int  patno;
    
    x2 = 0;	
    y2 = py;

    for (y=0; y<toy; y++)
    { 
       x2 = px;
       for (x=0; x < tox; x++)
       {  patno = figure[y][x];
	  XPutImage(display, screenContext, graphContext, 
      	        PatIm[patno], 0, 0, x2, y2, FIG_WIDTH, FIG_HEIGHT);
          x2 += FIG_WIDTH;		
       }

       y2 += FIG_HEIGHT;
    }

    if (DoInvalidate)
    {  RECT rc;

       rc.top = py; rc.bottom = y2;
       rc.left = px; rc.right =  x2;

       ContextRectToScreen(&rc);
    }  
}


/*--------------------------------------------------------*/
void ShowShadow(FIGURE figure, int pos)
{  char shadow[BRD_WIDTH], s;
   int  x, y, x1, fromx, tox;
   RECT  rc;
   int col, row, patno;

   if (figure == NULL)
   {
      rc.top = ShadowY;  rc.bottom = ShadowY + FIG_HEIGHT;
      rc.left = BoardX;  rc.right = BoardX + BoardW;
      DrawSolidRect(&rc, 0);
      if (DoInvalidate) ContextRectToScreen(&rc); 
      return;
   }


   bzero (shadow, BRD_WIDTH);
   
   for (col=0; col<FIG_DIM; col++)
   {	x1 = pos + col;
	if (x1 >= BRD_WIDTH) break;

	s = 0;
	for (row=0; row<FIG_DIM; row++)
	{ patno = figure[row][col];
	   if (patno)
           { s = patno; break; }
	}	

	shadow[x1] = s;

    }



   fromx = pos-1;
   if (fromx < 0)  fromx = 0;
   tox = pos + FIG_DIM +1 ;
   if (tox > BRD_WIDTH) tox = BRD_WIDTH;
//   fromx = 0;
//   tox = BRD_WIDTH;
    x = BoardX + fromx * FIG_WIDTH;
    y = ShadowY;

    rc.left  = x;


    for (col=fromx; col<tox; col++)
    {  patno = shadow[col];

       XPutImage(display, screenContext, graphContext, 
      	        PatIm[patno], 0, 0, x, y, FIG_WIDTH, FIG_HEIGHT);  
       x += FIG_WIDTH;		
    }

    rc.right = x;
     
    rc.top   = y;
    rc.bottom= y + FIG_HEIGHT;
    
    if (DoInvalidate)  ContextRectToScreen(&rc);
}

//====================================================================

static void ShowJoyArrow(int arrowtype,  XImage *JoyArrIm[])
{   int	x, y;

    x =  JoyEX + ArrOffs[arrowtype][0];
    y =  JoyEY + ArrOffs[arrowtype][1];

    XPutImage(display, screenContext, graphContext, 
          JoyArrIm[arrowtype], 0, 0, x, y, JOYARR_WIDTH, JOYARR_HEIGHT);
/*
    if (DoInvalidate)
    {  RECT rc;
       rc.left = x;   rc.top = y;
       rc.right = x + JOYARR_WIDTH;
       rc.bottom = x + JOYARR_HEIGHT;
       ContextRectToScreen(&rc);
    }                  
*/    
}

void  ShowJoyPos(int active)
{   int i;

     for (i=0; i<JOYARR_COUNT; i++)
        ShowJoyArrow(i, (i==active) ? JoyArrActiveIm : JoyArrPassiveIm);

    if (DoInvalidate)
    {  RECT rc;
       rc.left = JoyEX; rc.top = JoyEY;
       rc.right = JoyEX + JoyW - 8;
       rc.bottom = JoyEY + JoyH - 4;
       ContextRectToScreen(&rc);
    }                  

}

//====================================================================

const char *WindowIdAtomName = "XITETRIS_WINDOW_ID";
static Atom WindowIdAtom;
static Bool IsAppWindow(Window RootWnd, Window QuestionedWnd)
{	Bool ret = False;
	Window *WndPtr, WndTemp;
	unsigned int WndCount;  

  
  	if (XQueryTree(display, RootWnd, &WndTemp, &WndTemp, &WndPtr, &WndCount))
	{	while (WndCount > 0 && *WndPtr != QuestionedWnd)
		{	WndPtr++; WndCount--; }
		
		ret = (WndCount > 0);	
	
	}

	return ret;
}

static Bool GetPausedWindow(Window RootWnd, Window *PausedWndPtr, Bool DeleteFlag)
{ 
    Atom actual_type;
    int actual_format;
    unsigned long nitems, bytes_after;
	Window PausedWnd;
    unsigned char *data;
       	
    WindowIdAtom = XInternAtom(display, WindowIdAtomName,  False); 
    if (WindowIdAtom == None) return False;
    

    XGetWindowProperty(display, RootWnd, WindowIdAtom, 0,
			     1, DeleteFlag, XA_WINDOW,
			     &actual_type,  &actual_format,
			     &nitems, &bytes_after, &data); 

    
    if (actual_type == None || nitems == 0) return False;
    
    PausedWnd = *(long *)data;	    
    XFree(data);

	if (!IsAppWindow(RootWnd,  PausedWnd)) return False;

	*PausedWndPtr = PausedWnd;
	return True;
	
}


/*
void ClearPausedWindow(void)
{

	Window PausedWnd;
	
    Window RootWnd = RootWindow(display, screenNumber);

	if (GetPausedWindow(RootWnd, &PausedWnd, False) &&
		PausedWnd == window)
		XDeleteProperty(display, RootWnd, WindowIdAtom);
	
}
*/

static int ResumePausedWindow(Window RootWnd, char *message, int mode)
{
    Window PausedWnd;
    XEvent xevnt;


    if (!GetPausedWindow(RootWnd, &PausedWnd, True))
    {  if (message) strcpy (message, "No paused windows");
       return 8;
    }
    
/*
    xevnt.xclient.type = ClientMessage;
    xevnt.xclient.serial = 0;
    xevnt.xclient.send_event = True;
    xevnt.xclient.message_type = WindowIdAtom;           
    xevnt.xclient.data.l[0] = (long) mode;
    xevnt.xclient.display = display;
*/

  
    

    xevnt.xkey.type = KeyPress;
    xevnt.xkey.display = display;
    xevnt.xkey.send_event = True;
    xevnt.xkey.keycode = mode;

     
    XSendEvent(display, PausedWnd,  False, KeyPressMask, &xevnt);   

    return 0;
}

void PauseGame(void)
{ 
    extern char  *PlayListFileName;
    extern int   MusicEnabled;
    extern int   Confined;

    
    Window RootWnd = RootWindow(display, screenNumber);
    XEvent xevnt;
    int  OldMusicEnabled;


    ResumePausedWindow(RootWnd, NULL, 0);

    
    WindowIdAtom = XInternAtom(display, WindowIdAtomName,  False); 
    if (WindowIdAtom == None) return;

   
    XChangeProperty(display, RootWnd, WindowIdAtom, XA_WINDOW, 32,
			    PropModeReplace, (unsigned char *) &window, 1);

    OldMusicEnabled = MusicEnabled;
    
    if (PlayListFileName)
    {
       MusicEnabled = 0;
       PrepareMusic(MODE_FORCE);
    }                   

    if (Confined) ReleasePointer();
    
 /*    SuspendTimer();  */
    			    
    XUnmapWindow(display, window);			    


    do   XNextEvent(display, &xevnt);
    while  (xevnt.type != KeyPress  ||
	    !xevnt.xkey.send_event);

  /*    ResumeTimer();  */

    if (xevnt.xkey.keycode == 0) 
    {  TerminationHandler(8);
       return;
    }   

    XMapWindow(display, window);

   if (PlayListFileName)
   {  MusicEnabled = OldMusicEnabled;
      PrepareMusic(MODE_FORCE);
   }     
       			     
}			       

/* Mode = 0 - resume, 1 - kill */
int ResumeGame(char *DispName, char *message, int mode)
{   
    Window RootWnd;
    
    display = XOpenDisplay(DispName);
    if (!display)
    {  strcpy(message, "Could not open display ");
       if (DispName) strcat(message, DispName);
       return 8;
    }     

    RootWnd = RootWindow(display, DefaultScreen(display));
    return ResumePausedWindow(RootWnd, message, mode);
} 
    		


extern Atom WMStateAtom;

long  GetWindowState()	
{       		
    Atom actual_type;
    int actual_format;
    unsigned long nitems, bytes_after;
    unsigned char *data;
    long state;

    XGetWindowProperty(display, window, WMStateAtom, 0,
			     2, False, WMStateAtom,
			     &actual_type,  &actual_format,
			     &nitems, &bytes_after, &data); 
			     
    if (nitems > 0) state = *(long *)data; else state = -1;
    XFree(data);
    return state;    			     
}


void SuspendGame(void)
{
    while (GetWindowState() == IconicState) 
    {  if (!TimerSuspended) pause();
       kbdGet();
    }   

}
